# 陈茂森论文

## 随机移动网络系统的稳定性

### 马尔科夫链与网络平均度推导

**1.马尔科夫链的基本概念**

马尔科夫链描述的是这样一种随机过程：系统在若干个可能的状态中变化，**下一时刻所处状态只依赖于当前状态**，而与过去的状态无关，这就是所谓的“无记忆性”或**马尔科夫性**。

**无记忆性**意味着，对于任何 $s, t \ge 0$，
$$
P(T > s+t \mid T > s) = P(T > t).
$$

假设你已经等待了 $s$ 分钟，那么再等待至少 $t$ 分钟的概率，和你一开始就等待至少 $t$ 分钟的概率完全相同。



在这个模型中，每条链路只有两个可能的状态：

- **状态0**：链路断开
- **状态1**：链路连通

设在时刻 $t$ 时，某条链路处于连通状态的概率为 $p_1(t)$；由于只有两种状态，所以断开的概率就是  
$$
p_0(t)=1-p_1(t).
$$

同时，我们假设链路从一个状态转移到另一个状态需要等待一段时间，这段**等待时间**通常服从**指数分布**（论文中通过 KS 检验确认）。这意味着，从0到1和从1到0有两个转移速率，我们记作：

- 从0到1的转移速率为 $\lambda_{01}$  
- 从1到0的转移速率为 $\lambda_{10}$

这些速率表示单位时间内发生状态转换的可能性。

**2.推导单条链路的连通概率**

根据连续时间马尔科夫链的理论，我们可以写出**状态转移的微分方程**。对于状态1（连通状态），概率 $p_1(t)$ 的变化率由两个部分组成：

1. 当链路处于状态0时，以速率 $\lambda_{01}$ 变为状态1。这部分概率增加的速率为  
   $$
   \lambda_{01} \, p_0(t)=\lambda_{01} (1-p_1(t)).
   $$

2. 当链路处于状态1时，以速率 $\lambda_{10}$ 转换为状态0。这部分使 $p_1(t)$ 减少，其速率为  
   $$
   \lambda_{10} \, p_1(t).
   $$

所以，$p_1(t)$ 的微分方程写成：
$$
\frac{d p_1(t)}{dt} = \lambda_{01} \, (1-p_1(t)) - \lambda_{10} \, p_1(t).
$$

这个方程可以整理为：
$$
\frac{d p_1(t)}{dt} + (\lambda_{01}+\lambda_{10}) \, p_1(t) = \lambda_{01}.
$$

这其实是一个一阶线性微分方程，其标准求解方法是求解其齐次解与非齐次解。

**3. 求解微分方程**

整个微分方程的通解为：
$$
p_1(t)= \frac{\lambda_{01}}{\lambda_{01}+\lambda_{10}} + C\, e^{-(\lambda_{01}+\lambda_{10})t}.
$$

利用初始条件 $p_1(0)=p_1^0$（初始时刻链路连通的概率），我们可以求出 $C$：

即
$$
C = p_1^0 -\frac{\lambda_{01}}{\lambda_{01}+\lambda_{10}}.
$$

所以，链路在任意时刻 $t$ 连通的概率为：
$$
p_1(t)= \frac{\lambda_{01}}{\lambda_{01}+\lambda_{10}} + \left( p_1^0 -\frac{\lambda_{01}}{\lambda_{01}+\lambda_{10}} \right)e^{-(\lambda_{01}+\lambda_{10})t}.
$$

这就是单条链路的连通概率函数，描述了从任意初始条件出发，经过一段时间后，链路达到平衡状态的过程。

**4.推导网络平均度的变化函数**

在一个由 $N$ 个节点构成的网络中，每个节点都与其它节点进行通信（不考虑自环），因此每个节点最多有 $N-1$ 个邻居。对于任意一对节点 $i$ 和 $j$，它们之间链路连通的概率 $p_1(t)$（假设所有链路**独立且同分布**）。

- 某个节点 $i$ 在时刻 $t$ 的度 $d_i(t)$可以写作：
  $$
  d_i(t)= \sum_{\substack{j=1 \\ j\neq i}}^N p_{ij}(t),
  $$
  其中 $p_{ij}(t)=p_1(t)$。

- 因此，每个节点的期望度为：
  $$
  E[d_i(t)]=(N-1)p_1(t).
  $$

- 网络平均度就是对所有节点的期望度取平均，由于网络中每个节点都遵循相同统计规律，所以网络平均度可表示为：
  $$
  \bar{d}(t) = \frac{1}{N}\sum_{i=1}^{N} E[d_i(t)] = \frac{N \cdot (N-1)p_1(t)}{N} = (N-1)p_1(t)
  $$

将我们前面得到的 $p_1(t)$ 表达式代入，就得到网络平均度随时间变化的表达式：
$$
\bar{d}(t)= (N-1)\left[ \frac{\lambda_{01}}{\lambda_{01}+\lambda_{10}} + \left( p_1^0 -\frac{\lambda_{01}}{\lambda_{01}+\lambda_{10}} \right)e^{-(\lambda_{01}+\lambda_{10})t}\right].
$$

这就是**网络平均度的变化函数**：

- 网络开始时每条链路的连通概率为 $p_1^0$ 
- 这种调整过程符合指数衰减规律，即偏离平衡值的部分按 $e^{-(\lambda_{01}+\lambda_{10})t}$ 衰减
- 当 $t$ 趋向无穷大时，指数项 $e^{-(\lambda_{01}+\lambda_{10})t}$ 衰减为0，网络平均度趋向于  $(N-1)\frac{\lambda_{01}}{\lambda_{01}+\lambda_{10}}$，这就是网络达到平衡状态后的理论平均度。



### 特征信号参数的平稳性

证明**系统在平衡态下具有统计上的稳定性**。

#### **从节点空间分布证明平稳性。**

设节点在模型区域的坐标为 $(X,Y)$，其分布概率密度函数写为
$$
f(x,y).
$$

那么节点在模型子区域 $R_1$ 中出现的概率为

$$
P_{R_1}=\int_{R_1} f(x,y) \,dx\,dy.
$$

在平衡状态下，理论上节点的位置分布 **$f(x,y)$ 保持不变**，即每个区域内节点出现的概率 $P_{R_1}$ 是常数，不随时间变化。证明在平衡状态下节点分布稳定。

#### **扰动后的恢复能力**

- 静止节点分布特性满足均匀分布,概率密度函数为 $g(x,y)$；
- 运动节点的概率密度函数为 $h(x,y)$；
- **在时刻 $t_0$ 时**，网络中共有 $N$ 个节点，其中有 $s$ 个静止，故**静止节点的比例为 $p=\frac{s}{N}$**。

节点整体的分布概率密度函数可写为

$$
f(x,y)=p\, g(x,y)+(1-p)\, h(x,y).
$$

在平衡状态下，$p$ 的理论值为一个常数，所以 $f(x,y)$ 不随时间变化，从而网络连通度稳定。



接下来，**考虑外界扰动**的影响：假设在**时刻 $t_1$** 新加入 $m$ 个**符合均匀分布的节点**，

**扰动后的总分布（$t_1$时刻后）**

- 新加入的 $m$ 个节点是静止的，其分布为 $g(x,y)$
- 此时网络的总节点数 $N+m$：
  - **静止节点总数**：$s+m$
  - **运动节点总数**：$N-s$（原有运动节点数不变）

因此，扰动后的分布为：  
$$
f(x,y,t_1) = \frac{s + m}{N + m} g(x,y) + \frac{N - s}{N + m} h(x,y)
$$

$$
f(x,y,t_1) = p' \cdot g(x,y) + (1-p') \cdot h(x,y)
$$

其中 $p' = \frac{s + m}{N + m}$。

**近似处理（当 $N, s \gg m$ 时）**
$$
p' = \frac{s + m}{N + m} \approx \frac{s}{N} = p
$$
因此，扰动后的分布近似为：  
$$
f(x,y,t_1) \approx p \cdot g(x,y) + (1-p) \cdot h(x,y)
$$
这与初始平衡态的分布相同，说明网络在扰动后恢复了平衡态。



### 系统稳定性分析

**建立系统状态方程与平衡点**

论文将随机移动网络的动态演化描述为一个一般的状态方程：

$$
\frac{dx}{dt} = f(x, t)
$$
其中，$x$ 是系统的 $n$ 维状态向量，$f(x, t)$ 是描述状态随时间变化的函数。

- **平衡点定义**: 当存在一个状态 $x_e$ 满足对任意 $t$，有
  $f(x_e, t) = 0$
  这时 $x_e$ 就是系统的平衡状态。如论文中特别关注的网络平均度 $x_d$，不再随时间变化。

**采用李雅普诺夫第二类方法**

由于本系统状态向量各分量间关系复杂，且无法求出状态矩阵的全部特征值，所以不能采用第一类方法。因此选择构造“李雅普诺夫函数”（第二类方法）来验证系统的稳定性。

**构造李雅普诺夫函数**
$$
V(x) = (x - x_e)^T P (x - x_e)
$$
其中 $P$ 是一个正定矩阵。由此保证：

- **正定性**: 对于除 $x = x_e$ 外的所有状态，$V(x) > 0$；且在平衡点 $x_e$ 处有 $V(x_e) = 0$。

**分析李雅普诺夫函数的时间导数**
$$
\dot{V}(x) = \frac{\partial V(x)}{\partial x} \cdot f(x, t)
$$
**平衡时 $\dot{V}(x) = 0$**: 当且仅当系统处于平衡状态 $x = x_e$ 时，有 $\dot{V}(x) = 0$。

同时在平衡附近的非平衡状态下，由于选定的李雅普诺夫函数“能量”不会增加，从而得到$\dot{V}(x) ≤ 0$
